# line 1 "o/character.d"

# line 1 "o/character.d"
/* -*-C-*- */
/*
 Copyright (C) 1994 M. Hagiya, W. Schelter, T. Yuasa
 Copyright (C) 2024 Camm Maguire

This file is part of GNU Common Lisp, herein referred to as GCL

GCL is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GCL is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public 
License for more details.

You should have received a copy of the GNU Library General Public License 
along with GCL; see the file COPYING.  If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
	character.d

	character routines
*/

#include "include.h"

#define	CHFONTLIM	1	/*  character font limit  */
#define	CHBITSLIM	1	/*  character bits limit  */
#define	CHCODEFLEN	8	/*  character code field length  */
#define	CHFONTFLEN	0	/*  character font field length  */
#define	CHBITSFLEN      0	/*  character bits field length  */

void
Lstandard_char_p()
{
# line 37 "o/character.d"

	int i;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 39 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 42 "o/character.d"

	i = char_code(c);
	if ((' ' <= i && i < '\177') || i == '\n')
		{
			vs_base[0] = Ct;
			vs_top = vs_base + 1;
			return;
		}
# line 45 "o/character.d"

	{
		vs_base[0] = Cnil;
		vs_top = vs_base + 1;
		return;
	}
# line 46 "o/character.d"

#undef c
}
# line 47 "o/character.d"


void
Lgraphic_char_p()
{
# line 49 "o/character.d"

	int i;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 51 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 54 "o/character.d"

	i = char_code(c);
	if (' ' <= i && i < '\177')
		{
			vs_base[0] = Ct;
			vs_top = vs_base + 1;
			return;
		}
# line 57 "o/character.d"

	{
		vs_base[0] = Cnil;
		vs_top = vs_base + 1;
		return;
	}
# line 58 "o/character.d"

#undef c
}
# line 59 "o/character.d"


void
Lalpha_char_p()
{
# line 61 "o/character.d"

	int i;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 63 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 66 "o/character.d"

	i = char_code(c);
	if (isalpha(i))
		{
			vs_base[0] = Ct;
			vs_top = vs_base + 1;
			return;
		}
# line 69 "o/character.d"

	else
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 71 "o/character.d"

#undef c
}
# line 72 "o/character.d"


void
Lupper_case_p()
{
# line 74 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 75 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 78 "o/character.d"

	if (isUpper(char_code(c)))
		{
			vs_base[0] = Ct;
			vs_top = vs_base + 1;
			return;
		}
# line 80 "o/character.d"

	{
		vs_base[0] = Cnil;
		vs_top = vs_base + 1;
		return;
	}
# line 81 "o/character.d"

#undef c
}
# line 82 "o/character.d"


void
Llower_case_p()
{
# line 84 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 85 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 88 "o/character.d"

	if (isLower(char_code(c)))
		{
			vs_base[0] = Ct;
			vs_top = vs_base + 1;
			return;
		}
# line 90 "o/character.d"

	{
		vs_base[0] = Cnil;
		vs_top = vs_base + 1;
		return;
	}
# line 91 "o/character.d"

#undef c
}
# line 92 "o/character.d"


void
Lboth_case_p()
{
# line 94 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 95 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 98 "o/character.d"

	if (isUpper(char_code(c)) || isLower(char_code(c)))
		{
			vs_base[0] = Ct;
			vs_top = vs_base + 1;
			return;
		}
# line 100 "o/character.d"

	else
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 102 "o/character.d"

#undef c
}
# line 103 "o/character.d"


/*
	Digitp(i, r) returns the weight of code i
	as a digit of radix r.
	If r > 36 or i is not a digit, -1 is returned.
*/
int
digitp(i, r)
int i, r;
{
	if ('0' <= i && i <= '9' && 1 < r && i < '0' + r)
		return(i - '0');
	if ('A' <= i && 10 < r && r <= 36 && i < 'A' + (r - 10))
		return(i - 'A' + 10);
	if ('a' <= i && 10 < r && r <= 36 && i < 'a' + (r - 10))
		return(i - 'a' + 10);
	return(-1);
}

void
Ldigit_char_p()
{
# line 123 "o/character.d"

	int d;
	int narg;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]
#define	r	DPPbase[1+0]

	narg = vs_top - vs_base;
	if (narg < 1)
		too_few_arguments();
	if (narg <= 1 + 0) {
		vs_push(make_fixnum(10));
		narg++;
	}
	if (narg > 1 + 1)
		too_many_arguments();

# line 125 "o/character.d"

	check_type_character(&c);
	check_type_non_negative_integer(&r);
	if (type_of(r) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 129 "o/character.d"

	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 131 "o/character.d"

	d = digitp(char_code(c), fix(r));
	if (d < 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 134 "o/character.d"

	{
		vs_base[0] = make_fixnum(d);
		vs_top = vs_base + 1;
		return;
	}
# line 135 "o/character.d"

#undef c
#undef r
}
# line 136 "o/character.d"


void
Lalphanumericp()
{
# line 138 "o/character.d"

	int i;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 140 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 143 "o/character.d"

	i = char_code(c);
	if (isalphanum(i))
		{
			vs_base[0] = Ct;
			vs_top = vs_base + 1;
			return;
		}
# line 146 "o/character.d"

	else
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 148 "o/character.d"

#undef c
}
# line 149 "o/character.d"


bool
char_eq(x, y)
object x, y;
{
	return(char_code(x) == char_code(y)
	    && char_bits(x) == char_bits(y)
	    && char_font(x) == char_font(y));
}

void
Lchar_eq()
{
# line 160 "o/character.d"

	int i;
	int narg;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	narg = vs_top - vs_base;
	if (narg < 1)
		too_few_arguments();

# line 162 "o/character.d"

	check_type_character(&c);
	for (i = 0;  i < narg;  i++)
		check_type_character(&vs_base[i]);
	for (i = 1;  i < narg;  i++)
		if (!char_eq(vs_base[i-1], vs_base[i]))
			{
				vs_base[0] = Cnil;
				vs_top = vs_base + 1;
				return;
			}
# line 168 "o/character.d"

	{
		vs_base[0] = Ct;
		vs_top = vs_base + 1;
		return;
	}
# line 169 "o/character.d"

#undef c
}
# line 170 "o/character.d"


void
Lchar_neq()
{
# line 172 "o/character.d"

	int i, j;
	int narg;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	narg = vs_top - vs_base;
	if (narg < 1)
		too_few_arguments();

# line 174 "o/character.d"

	check_type_character(&c);
	for (i = 0;  i < narg;  i++)
		check_type_character(&vs_base[i]);
	if (narg == 0)
		{
			vs_base[0] = Ct;
			vs_top = vs_base + 1;
			return;
		}
# line 179 "o/character.d"

	for (i = 1;  i < narg;  i++)
		for (j = 0;  j < i;  j++)
			if (char_eq(vs_base[j], vs_base[i]))
				{
					vs_base[0] = Cnil;
					vs_top = vs_base + 1;
					return;
				}
# line 183 "o/character.d"

	{
		vs_base[0] = Ct;
		vs_top = vs_base + 1;
		return;
	}
# line 184 "o/character.d"

#undef c
}
# line 185 "o/character.d"



static int
char_cmp(x, y)
object x, y;
{
	if (char_font(x) < char_font(y))
		return(-1);
	if (char_font(x) > char_font(y))
		return(1);
	if (char_bits(x) < char_bits(y))
		return(-1);
	if (char_bits(x) > char_bits(y))
		return(1);
	if (char_code(x) < char_code(y))
		return(-1);
	if (char_code(x) > char_code(y))
		return(1);
	return(0);
}

static void
Lchar_cmp(s, t)
int s, t;
{
	int narg, i;

	narg = vs_top - vs_base;
	if (narg == 0)
		too_few_arguments();
	for (i = 0; i < narg; i++)
		check_type_character(&vs_base[i]);
	for (i = 1; i < narg; i++)
		if (s*char_cmp(vs_base[i], vs_base[i-1]) < t) {
			vs_top = vs_base+1;
			vs_base[0] = Cnil;
			return;
		}
	vs_top = vs_base+1;
	vs_base[0] = Ct;
}

LFD(Lchar_l)()  { Lchar_cmp( 1, 1); }
LFD(Lchar_g)()  { Lchar_cmp(-1, 1); }
LFD(Lchar_le)() { Lchar_cmp( 1, 0); }
LFD(Lchar_ge)() { Lchar_cmp(-1, 0); }


bool
char_equal(x, y)
object x, y;
{
	int i, j;

	i = char_code(x);
	j = char_code(y);
	if (isLower(i))
		i -= 'a' - 'A';
	if (isLower(j))
		j -= 'a' - 'A';
	return(i == j);
}

void
Lchar_equal()
{
# line 249 "o/character.d"

	int i;
	int narg;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	narg = vs_top - vs_base;
	if (narg < 1)
		too_few_arguments();

# line 251 "o/character.d"

	check_type_character(&c);
	for (i = 0;  i < narg;  i++)
		check_type_character(&vs_base[i]);
	for (i = 1;  i < narg;  i++)
		if (!char_equal(vs_base[i-1], vs_base[i]))
			{
				vs_base[0] = Cnil;
				vs_top = vs_base + 1;
				return;
			}
# line 257 "o/character.d"

	{
		vs_base[0] = Ct;
		vs_top = vs_base + 1;
		return;
	}
# line 258 "o/character.d"

#undef c
}
# line 259 "o/character.d"


void
Lchar_not_equal()
{
# line 261 "o/character.d"

	int i, j;
	int narg;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	narg = vs_top - vs_base;
	if (narg < 1)
		too_few_arguments();

# line 263 "o/character.d"

	check_type_character(&c);
	for (i = 0;  i < narg;  i++)
		check_type_character(&vs_base[i]);
	for (i = 1;  i < narg;  i++)
		for (j = 0;  j < i;  j++)
			if (char_equal(vs_base[j], vs_base[i]))
				{
					vs_base[0] = Cnil;
					vs_top = vs_base + 1;
					return;
				}
# line 270 "o/character.d"

	{
		vs_base[0] = Ct;
		vs_top = vs_base + 1;
		return;
	}
# line 271 "o/character.d"

#undef c
}
# line 272 "o/character.d"



static int
char_compare(x, y)
object x, y;
{
	int i, j;

	i = char_code(x);
	j = char_code(y);
	if (isLower(i))
		i -= 'a' - 'A';
	if (isLower(j))
		j -= 'a' - 'A';
	if (i < j)
		return(-1);
	else if (i == j)
		return(0);
	else
		return(1);
}

static void
Lchar_compare(s, t)
int s, t;
{
	int narg, i;

	narg = vs_top - vs_base;
	if (narg == 0)
		too_few_arguments();
	for (i = 0; i < narg; i++)
		check_type_character(&vs_base[i]);
	for (i = 1; i < narg; i++)
		if (s*char_compare(vs_base[i], vs_base[i-1]) < t) {
			vs_top = vs_base+1;
			vs_base[0] = Cnil;
			return;
		}
	vs_top = vs_base+1;
	vs_base[0] = Ct;
}

LFD(Lchar_lessp)()        { Lchar_compare( 1, 1); }
LFD(Lchar_greaterp)()     { Lchar_compare(-1, 1); }
LFD(Lchar_not_greaterp)() { Lchar_compare( 1, 0); }
LFD(Lchar_not_lessp)()    { Lchar_compare(-1, 0); }


object
coerce_to_character(x)
object x;
{
BEGIN:
	switch (type_of(x)) {
	case t_fixnum:
		if (0 <= fix(x) && fix(x) < CHCODELIM)
			return(code_char(fix(x)));
		break;

	case t_character:
		return(x);

	case t_symbol:
	  x=coerce_to_string(x);
	case t_simple_string:
	case t_string:
	  if (VLEN(x) == 1)
	    return(code_char(x->ust.ust_self[0]));
		break;
	default:
		break;
	}
	vs_push(x);
	x = wrong_type_argument(sLcharacter, x);
	vs_popp;
	goto BEGIN;
}

void
Lcharacter()
{
# line 352 "o/character.d"

	register object *DPPbase=vs_base;
#define	x	DPPbase[0]

	check_arg(1);

# line 353 "o/character.d"

	{
		vs_base[0] = coerce_to_character(x);
		vs_top = vs_base + 1;
		return;
	}
# line 354 "o/character.d"

#undef x
}
# line 355 "o/character.d"


void
Lchar_code()
{
# line 357 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 358 "o/character.d"

	check_type_character(&c);
	{
		vs_base[0] = make_fixnum(char_code(c));
		vs_top = vs_base + 1;
		return;
	}
# line 360 "o/character.d"

#undef c
}
# line 361 "o/character.d"


void
Lcode_char()
{
# line 363 "o/character.d"

	object x;
	int narg;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]
#define	b	DPPbase[1+0]
#define	f	DPPbase[1+1]

	narg = vs_top - vs_base;
	if (narg < 1)
		too_few_arguments();
	if (narg <= 1 + 0) {
		vs_push(make_fixnum(0));
		narg++;
	}
	if (narg <= 1 + 1) {
		vs_push(make_fixnum(0));
		narg++;
	}
	if (narg > 1 + 2)
		too_many_arguments();

# line 365 "o/character.d"

	check_type_non_negative_integer(&c);
        b=make_fixnum(0);/*FIXME*/
	check_type_non_negative_integer(&b);
        f=make_fixnum(0);/*FIXME*/
	check_type_non_negative_integer(&f);
	if (type_of(c) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 372 "o/character.d"

	if (type_of(b) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 374 "o/character.d"

	if (type_of(f) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 376 "o/character.d"

	if (fix(c)>=CHCODELIM || fix(b)>=CHBITSLIM || fix(f)>=CHFONTLIM)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 378 "o/character.d"

	if (fix(b) == 0 && fix(f) == 0)
		{
			vs_base[0] = code_char(fix(c));
			vs_top = vs_base + 1;
			return;
		}
# line 380 "o/character.d"

	x = alloc_object(t_character);
	char_code(x) = fix(c);
	char_bits(x) = fix(b);
	char_font(x) = fix(f);
	{
		vs_base[0] = x;
		vs_top = vs_base + 1;
		return;
	}
# line 385 "o/character.d"

#undef c
#undef b
#undef f
}
# line 386 "o/character.d"


void
Lchar_upcase()
{
# line 388 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 389 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = c;
			vs_top = vs_base + 1;
			return;
		}
# line 392 "o/character.d"

	if (isLower(char_code(c)))
		{
			vs_base[0] = code_char(char_code(c) - ('a' - 'A'));
			vs_top = vs_base + 1;
			return;
		}
# line 394 "o/character.d"

	else
		{
			vs_base[0] = c;
			vs_top = vs_base + 1;
			return;
		}
# line 396 "o/character.d"

#undef c
}
# line 397 "o/character.d"


void
Lchar_downcase()
{
# line 399 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 400 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 403 "o/character.d"

	if (isUpper(char_code(c)))
		{
			vs_base[0] = code_char(char_code(c) + ('a' - 'A'));
			vs_top = vs_base + 1;
			return;
		}
# line 405 "o/character.d"

	else
		{
			vs_base[0] = c;
			vs_top = vs_base + 1;
			return;
		}
# line 407 "o/character.d"

#undef c
}
# line 408 "o/character.d"


int
digit_weight(w, r)
int w, r;
{
	if (r < 2 || r > 36 || w < 0 || w >= r)
		return(-1);
	if (w < 10)
		return(w + '0');
	else
		return(w - 10 + 'A');
}

void
Ldigit_char()
{
# line 422 "o/character.d"

	object x,f;
	int dw;
	int narg;
	register object *DPPbase=vs_base;
#define	w	DPPbase[0]
#define	r	DPPbase[1+0]

	narg = vs_top - vs_base;
	if (narg < 1)
		too_few_arguments();
	if (narg <= 1 + 0) {
		vs_push(make_fixnum(10));
		narg++;
	}
	if (narg > 1 + 1)
		too_many_arguments();

# line 425 "o/character.d"

	check_type_non_negative_integer(&w);
	check_type_non_negative_integer(&r);
        f=make_fixnum(0);/*FIXME*/
	check_type_non_negative_integer(&f);
	if (type_of(w) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 431 "o/character.d"

	if (type_of(r) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 433 "o/character.d"

	if (type_of(f) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 435 "o/character.d"

	dw = digit_weight(fix(w), fix(r));
	if (dw < 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 438 "o/character.d"

	if (fix(f) >= CHFONTLIM)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 440 "o/character.d"

	if (fix(f) == 0)
		{
			vs_base[0] = code_char(dw);
			vs_top = vs_base + 1;
			return;
		}
# line 442 "o/character.d"

	x = alloc_object(t_character);
	char_code(x) = dw;
	char_bits(x) = 0;
	char_font(x) = fix(f);
	{
		vs_base[0] = x;
		vs_top = vs_base + 1;
		return;
	}
# line 447 "o/character.d"

#undef w
#undef r
}
# line 448 "o/character.d"


void
Lchar_int()
{
# line 450 "o/character.d"

	int i;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 452 "o/character.d"

	check_type_character(&c);
	i = (char_font(c)*CHBITSLIM + char_bits(c))*CHCODELIM
	  + char_code(c);
	{
		vs_base[0] = make_fixnum(i);
		vs_top = vs_base + 1;
		return;
	}
# line 456 "o/character.d"

#undef c
}
# line 457 "o/character.d"


void
Lchar_name()
{
# line 459 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 460 "o/character.d"

	check_type_character(&c);
	if (char_bits(c) != 0 || char_font(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 463 "o/character.d"

	switch (char_code(c)) {
	case '\r':
		{
			vs_base[0] = STreturn;
			vs_top = vs_base + 1;
			return;
		}
# line 466 "o/character.d"


	case ' ':
		{
			vs_base[0] = STspace;
			vs_top = vs_base + 1;
			return;
		}
# line 469 "o/character.d"


	case '\177':
		{
			vs_base[0] = STrubout;
			vs_top = vs_base + 1;
			return;
		}
# line 472 "o/character.d"

	
	case '\f':
		{
			vs_base[0] = STpage;
			vs_top = vs_base + 1;
			return;
		}
# line 475 "o/character.d"


	case '\t':
		{
			vs_base[0] = STtab;
			vs_top = vs_base + 1;
			return;
		}
# line 478 "o/character.d"


	case '\b':
		{
			vs_base[0] = STbackspace;
			vs_top = vs_base + 1;
			return;
		}
# line 481 "o/character.d"


	case '\n':
		{
			vs_base[0] = STnewline;
			vs_top = vs_base + 1;
			return;
		}
# line 484 "o/character.d"

	}
        if (char_code(c)<' ' || char_code(c) >='\177') {
          object x=make_simple_string(" ");
          x->st.st_self[0]=char_code(c);
          {
	vs_base[0] = x;
	vs_top = vs_base + 1;
	return;
}
# line 489 "o/character.d"

        }
	{
		vs_base[0] = Cnil;
		vs_top = vs_base + 1;
		return;
	}
# line 491 "o/character.d"

#undef c
}
# line 492 "o/character.d"


void
Lname_char()
{
# line 494 "o/character.d"

	register object *DPPbase=vs_base;
#define	s	DPPbase[0]

	check_arg(1);

# line 495 "o/character.d"

	s = coerce_to_string(s);
	if (string_equal(s, STreturn))
		{
			vs_base[0] = code_char('\r');
			vs_top = vs_base + 1;
			return;
		}
# line 498 "o/character.d"

	if (string_equal(s, STspace))
		{
			vs_base[0] = code_char(' ');
			vs_top = vs_base + 1;
			return;
		}
# line 500 "o/character.d"

	if (string_equal(s, STrubout))
		{
			vs_base[0] = code_char('\177');
			vs_top = vs_base + 1;
			return;
		}
# line 502 "o/character.d"

	if (string_equal(s, STpage))
		{
			vs_base[0] = code_char('\f');
			vs_top = vs_base + 1;
			return;
		}
# line 504 "o/character.d"

	if (string_equal(s, STtab))
		{
			vs_base[0] = code_char('\t');
			vs_top = vs_base + 1;
			return;
		}
# line 506 "o/character.d"

	if (string_equal(s, STbackspace))
		{
			vs_base[0] = code_char('\b');
			vs_top = vs_base + 1;
			return;
		}
# line 508 "o/character.d"

	if (string_equal(s, STlinefeed) || string_equal(s, STnewline))
		{
			vs_base[0] = code_char('\n');
			vs_top = vs_base + 1;
			return;
		}
# line 510 "o/character.d"

        if (VLEN(s)==1) {
	vs_base[0] = code_char(s->st.st_self[0]);
	vs_top = vs_base + 1;
	return;
}
# line 511 "o/character.d"

        if (VLEN(s)==2 && s->st.st_self[0]=='^') {
	  int ch=s->st.st_self[1]-'A'+1;
	  {
		vs_base[0] = code_char(ch);
		vs_top = vs_base + 1;
		return;
	}
# line 514 "o/character.d"

        }
        if (VLEN(s)==3 && s->st.st_self[0]=='^' && s->st.st_self[1]=='\\' && s->st.st_self[2]=='\\') {
	  int ch=s->st.st_self[1]-'A'+1;
	  {
		vs_base[0] = code_char(ch);
		vs_top = vs_base + 1;
		return;
	}
# line 518 "o/character.d"

        }
        if (VLEN(s)==4 && s->st.st_self[0]=='\\') {
	  int ch=(s->st.st_self[1]-'0')*8*8+(s->st.st_self[2]-'0')*8+(s->st.st_self[3]-'0');
	  {
		vs_base[0] = code_char(ch);
		vs_top = vs_base + 1;
		return;
	}
# line 522 "o/character.d"

        }
	{
		vs_base[0] = Cnil;
		vs_top = vs_base + 1;
		return;
	}
# line 524 "o/character.d"

#undef s
}
# line 525 "o/character.d"


void
gcl_init_character()
{
	int i;

	for (i = 0;  i < CHCODELIM;  i++) {
	  object x=(object)(character_table+i),y=(object)(character_name_table+i);
	  set_type_of(x,t_character);
	  x->ch.ch_code = i;
	  x->ch.tt=((' ' <= i && i < '\177') || i == '\n');
	  x->ch.ch_font = 0;
	  x->ch.ch_bits = 0;
	  x->ch.ch_name=y;
	  set_type_of(y,t_simple_string);
	  y->sst.sst_hasfillp = FALSE;
	  y->sst.sst_adjustable = FALSE;
	  set_array_elttype(y,aet_ch);
	  y->sst.sst_rank = 1;
	  y->sst.sst_dim = 1;
	  y->sst.sst_self = (void *)&x->ch.ch_code;

	}

 	make_constant("CHAR-CODE-LIMIT", make_fixnum(CHCODELIM));
 	make_si_constant("CHAR-FONT-LIMIT", make_fixnum(CHFONTLIM));
 	make_si_constant("CHAR-BITS-LIMIT", make_fixnum(CHBITSLIM));

	STreturn = make_simple_string("Return");
	enter_mark_origin(&STreturn);
	STspace = make_simple_string("Space");
	enter_mark_origin(&STspace);
	STrubout = make_simple_string("Rubout");
	enter_mark_origin(&STrubout);
	STpage = make_simple_string("Page");
	enter_mark_origin(&STpage);
	STtab = make_simple_string("Tab");
	enter_mark_origin(&STtab);
	STbackspace = make_simple_string("Backspace");
	enter_mark_origin(&STbackspace);
	STlinefeed = make_simple_string("Linefeed");
	enter_mark_origin(&STlinefeed);

	STnewline = make_simple_string("Newline");
	enter_mark_origin(&STnewline);

	make_si_constant("CHAR-CONTROL-BIT", make_fixnum(0));
	make_si_constant("CHAR-META-BIT", make_fixnum(0));
	make_si_constant("CHAR-SUPER-BIT", make_fixnum(0));
	make_si_constant("CHAR-HYPER-BIT", make_fixnum(0));

}

void
Lmake_char()
{
# line 579 "o/character.d"

	object x;
	int code;
	int narg;
	register object *DPPbase=vs_base;
#define	c	DPPbase[0]
#define	b	DPPbase[1+0]
#define	f	DPPbase[1+1]

	narg = vs_top - vs_base;
	if (narg < 1)
		too_few_arguments();
	if (narg <= 1 + 0) {
		vs_push(make_fixnum(0));
		narg++;
	}
	if (narg <= 1 + 1) {
		vs_push(make_fixnum(0));
		narg++;
	}
	if (narg > 1 + 2)
		too_many_arguments();

# line 582 "o/character.d"

	check_type_character(&c);
	code = char_code(c);
	check_type_non_negative_integer(&b);
	check_type_non_negative_integer(&f);
	if (type_of(b) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 588 "o/character.d"

	if (type_of(f) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 590 "o/character.d"

	if (fix(b)>=CHBITSLIM || fix(f)>=CHFONTLIM)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 592 "o/character.d"

	if (fix(b) == 0 && fix(f) == 0)
		{
			vs_base[0] = code_char(code);
			vs_top = vs_base + 1;
			return;
		}
# line 594 "o/character.d"

	x = alloc_object(t_character);
	char_code(x) = code;
	char_bits(x) = fix(b);
	char_font(x) = fix(f);
	{
		vs_base[0] = x;
		vs_top = vs_base + 1;
		return;
	}
# line 599 "o/character.d"

#undef c
#undef b
#undef f
}
# line 600 "o/character.d"


void
Lchar_bits()
{
# line 602 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 603 "o/character.d"

	check_type_character(&c);
	{
		vs_base[0] = small_fixnum(char_bits(c));
		vs_top = vs_base + 1;
		return;
	}
# line 605 "o/character.d"

#undef c
}
# line 606 "o/character.d"


void
Lchar_font()
{
# line 608 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 609 "o/character.d"

	check_type_character(&c);
	{
		vs_base[0] = small_fixnum(char_font(c));
		vs_top = vs_base + 1;
		return;
	}
# line 611 "o/character.d"

#undef c
}
# line 612 "o/character.d"


void
Lchar_bit()
{
# line 614 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]
#define	n	DPPbase[1]

	check_arg(2);

# line 615 "o/character.d"

	check_type_character(&c);
	FEerror("Cannot get char-bit of ~S.", 1, c);
#undef c
#undef n
}
# line 618 "o/character.d"


void
Lset_char_bit()
{
# line 620 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]
#define	n	DPPbase[1]
#define	v	DPPbase[2]

	check_arg(3);

# line 621 "o/character.d"

	check_type_character(&c);
	FEerror("Cannot set char-bit of ~S.", 1, c);
#undef c
#undef n
#undef v
}
# line 624 "o/character.d"


void
Lstring_char_p()
{
# line 626 "o/character.d"

	register object *DPPbase=vs_base;
#define	c	DPPbase[0]

	check_arg(1);

# line 627 "o/character.d"

	check_type_character(&c);
	if (char_font(c) != 0 || char_bits(c) != 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 630 "o/character.d"

	{
		vs_base[0] = Ct;
		vs_top = vs_base + 1;
		return;
	}
# line 631 "o/character.d"

#undef c
}
# line 632 "o/character.d"


void
Lint_char()
{
# line 634 "o/character.d"

	int i, c, b, f;
	register object *DPPbase=vs_base;
#define	x	DPPbase[0]

	check_arg(1);

# line 636 "o/character.d"

	check_type_non_negative_integer(&x);
	if (type_of(x) == t_bignum)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 639 "o/character.d"

	i = fix(x);
	c = i % CHCODELIM;
	i /= CHCODELIM;
	b = i % CHBITSLIM;
	i /= CHBITSLIM;
	f = i % CHFONTLIM;
	i /= CHFONTLIM;
	if (i > 0)
		{
			vs_base[0] = Cnil;
			vs_top = vs_base + 1;
			return;
		}
# line 648 "o/character.d"

	if (b == 0 && f == 0)
		{
			vs_base[0] = code_char(c);
			vs_top = vs_base + 1;
			return;
		}
# line 650 "o/character.d"

	x = alloc_object(t_character);
	char_code(x) = c;
	char_bits(x) = b;
	char_font(x) = f;
	{
		vs_base[0] = x;
		vs_top = vs_base + 1;
		return;
	}
# line 655 "o/character.d"

#undef x
}
# line 656 "o/character.d"


void
gcl_init_character_function()
{
	make_function("STANDARD-CHAR-P", Lstandard_char_p);
	make_function("GRAPHIC-CHAR-P", Lgraphic_char_p);
	make_function("ALPHA-CHAR-P", Lalpha_char_p);
	make_function("UPPER-CASE-P", Lupper_case_p);
	make_function("LOWER-CASE-P", Llower_case_p);
	make_function("BOTH-CASE-P", Lboth_case_p);
	make_function("DIGIT-CHAR-P", Ldigit_char_p);
	make_function("ALPHANUMERICP", Lalphanumericp);
	make_function("CHAR=", Lchar_eq);
	make_function("CHAR/=", Lchar_neq);
	make_function("CHAR<", Lchar_l);
	make_function("CHAR>", Lchar_g);
	make_function("CHAR<=", Lchar_le);
	make_function("CHAR>=", Lchar_ge);
	make_function("CHAR-EQUAL", Lchar_equal);
	make_function("CHAR-NOT-EQUAL", Lchar_not_equal);
	make_function("CHAR-LESSP", Lchar_lessp);
	make_function("CHAR-GREATERP", Lchar_greaterp);
	make_function("CHAR-NOT-GREATERP", Lchar_not_greaterp);
	make_function("CHAR-NOT-LESSP", Lchar_not_lessp);
	make_function("CHARACTER", Lcharacter);
	make_function("CHAR-CODE", Lchar_code);
	make_function("CODE-CHAR", Lcode_char);
	make_function("CHAR-UPCASE", Lchar_upcase);
	make_function("CHAR-DOWNCASE", Lchar_downcase);
	make_function("DIGIT-CHAR", Ldigit_char);
	make_function("CHAR-INT", Lchar_int);
	make_function("CHAR-NAME", Lchar_name);
	make_function("NAME-CHAR", Lname_char);
	make_si_function("INT-CHAR", Lint_char);
	make_si_function("MAKE-CHAR", Lmake_char);
	make_si_function("CHAR-BITS", Lchar_bits);
	make_si_function("CHAR-FONT", Lchar_font);
	make_si_function("CHAR-BIT", Lchar_bit);
	make_si_function("SET-CHAR-BIT", Lset_char_bit);
	make_si_function("STRING-CHAR-P", Lstring_char_p);
}
